// Copyright (C) 2024 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

// ignore_for_file: avoid_equals_and_hash_code_on_mutable_classes

import 'package:sealed_result/sealed_result.dart' as sealed;

export 'future_or_result.dart';
export 'future_result.dart';

class ResultException extends sealed.ResultException {
  const ResultException(String super.message);
}

/// {@template result}
/// [Result] type is coming from functional languages where exceptions are
/// (rightfully) considered a side-effect, and therefore not appropriate to
/// pass domain errors. Mind the difference between different kinds of errors:
/// Some of them belong to domain, others don't.
///
/// *E.g. null reference exception or index out of bounds
/// are not related to domain - they rather indicate a defect.*
///
/// Either is defined as a generic type with two branches
/// - success with [Success]
/// - failure with [Failure]
///
/// It can appear in two forms, where
/// it contains an object of [Ok], or where it contains an object
/// of [Err]. It cannot appear in both states at once, or in none of them.
/// Therefore, if one possesses an [Result] instance, it either contains a
/// successfully produced result, or contains an error object.
/// {@endtemplate}
class Result<Success, Failure> {
  /// {@macro result}
  const Result._(this._result);

  factory Result.ok(Success value) => Ok(value);
  factory Result.err(Failure value) => Err(value);
  factory Result.error(Failure value) => Err(value);
  factory Result.success(Success value) => Ok(value);
  factory Result.failure(Failure value) => Err(value);

  final sealed.Result<Success, Failure> _result;

  /// Represents the left side of [Result] class.
  bool get isOk => _result.isOk;

  /// Represents the right side of [Result] class.
  bool get isErr => _result.isErr;

  /// Get nullable [Ok] value, and discarding the error, if any.
  Success? get ok => _result.ok;

  /// Get nullable [Err] value, and discarding the success value, if any.
  Failure? get err => _result.err;

  /// Get [U] value, may throw an exception.
  U unwrap<U>() {
    if (U == Success) {
      return fold<Success>(
        (value) => value,
        (right) => throw const ResultException(
          'Illegal use. You should check left value before calling',
        ),
      ) as U;
    }
    if (U == Failure) {
      return fold<Failure>(
        (left) => throw const ResultException(
          'Illegal use. You should check right value before calling',
        ),
        (value) => value,
      ) as U;
    }
    throw ResultException(
      'Illegal use. You should use $Success or $Failure type',
    );
  }

  /// Get **async** [U] value, may throw an exception.
  Future<U> unwrapAsync<U>() {
    if (U == Success) {
      return foldAsync<Success>(
        Future.value,
        (right) => throw const ResultException(
          'Illegal use. You should check left value before calling',
        ),
      ) as Future<U>;
    }
    if (U == Failure) {
      return foldAsync<Failure>(
        (left) => throw const ResultException(
          'Illegal use. You should check right value before calling',
        ),
        Future.value,
      ) as Future<U>;
    }
    throw ResultException(
      'Illegal use. You should use $Success or $Failure type',
    );
  }

  /// Fold [Ok] and [Err] into the value of one type
  U fold<U>(
    U Function(Success value) valueTransformer,
    U Function(Failure error) errorTransformer,
  ) =>
      _result.fold(
        (ok) => valueTransformer(ok),
        (err) => errorTransformer(err),
      );

  /// Fold [Ok] and [Err] **asynchronously** into the value of one type
  Future<U> foldAsync<U>(
    Future<U> Function(Success value) valueTransformer,
    Future<U> Function(Failure error) errorTransformer,
  ) =>
      _result.foldAsync(
        (ok) => valueTransformer(ok),
        (err) => errorTransformer(err),
      );

  /// Swap [Ok] and [Err]
  Result<Failure, Success> swap() => fold(
        Err.new,
        Ok.new,
      );

  /// Returns [res] if the [Result] is [Ok], otherwise returns
  /// the [Err] value of this.
  Result<U, Failure> and<U>(Result<U, Failure> res) => Result._(
        _result.and(res._result),
      );

  /// Returns [res] if the [Result] is [Err], otherwise returns
  /// the [Ok] value of this.
  Result<Success, F> or<F>(Result<Success, F> res) => Result._(
        _result.or(res._result),
      );

  /// Returns true if the result is an [Ok] or [Err] value containing
  /// the given value/error.
  bool contains<U>(U x) {
    if (x is Success) {
      return _result.contains(x);
    } else if (x is Failure) {
      return _result.containsErr(x);
    } else {
      throw ArgumentError('The value must be of type Success or Failure');
    }
  }

  /// Returns the contained [Ok] value. Throw [ResultException] on [Err]
  /// with its content.
  Success expect(String msg) {
    try {
      return _result.expect(msg);
    } on sealed.ResultException catch (e) {
      throw ResultException(e.message.toString());
    }
  }

  /// Returns the contained [Err] value. Throw [ResultException] on [Ok]
  /// with its content.
  Failure expectErr(String msg) {
    try {
      return _result.expectErr(msg);
    } on sealed.ResultException catch (e) {
      throw ResultException(e.message.toString());
    }
  }

  /// Maps a [Result<Success, Failure>] to [Result<U, Failure>] by applying a
  /// function to a contained [Ok] value, leaving an [Err] value untouched.
  Result<U, Failure> map<U>(U Function(Success value) mapper) =>
      Result._(_result.map(mapper));

  /// Maps a [Result<Success, Failure>] to [Result<U, Failure>] by applying
  /// an **async** function to a contained [Ok] value, leaving an [Err]
  /// value untouched.
  Future<Result<U, Failure>> mapAsync<U>(
    Future<U> Function(Success value) mapper,
  ) async {
    final result = await _result.mapAsync(mapper);
    return Result._(result);
  }

  /// Maps a [Result<Success, Failure>] to [Result<Success, F>] by applying a
  /// function to a contained [Err] value, leaving an [Ok] value untouched.
  Result<Success, F> mapErr<F>(F Function(Failure error) mapper) =>
      Result._(_result.mapErr(mapper));

  /// Maps a [Result<Success, Failure>] to [Result<U, Failure>] by applying
  /// an **async** function to a contained [Err] value, leaving an [Ok]
  /// value untouched.
  Future<Result<Success, F>> mapErrAsync<F>(
    Future<F> Function(Failure error) mapper,
  ) async {
    final result = await _result.mapErrAsync(mapper);
    return Result._(result);
  }

  /// Transforms a [Result<Success, Failure>] to [Result<U, F>] by applying
  /// functions to contained [Ok] and [Err] values.
  Result<U, F> either<U, F>(
    U Function(Success value) valueTransformer,
    F Function(Failure error) errorTransformer,
  ) =>
      Result._(
        _result
            .map(
              valueTransformer,
            )
            .mapErr(
              errorTransformer,
            ),
      );

  /// Transforms a [Result<Success, Failure>] to [Result<U, F>] by
  /// applying **async** functions to contained [Ok] and [Err] values.
  Future<Result<U, F>> eitherAsync<U, F>(
    Future<U> Function(Success value) valueTransformer,
    Future<F> Function(Failure error) errorTransformer,
  ) =>
      _result
          .mapAsync(
            valueTransformer,
          )
          .mapErrAsync(
            errorTransformer,
          )
          .then(Result._);

  /// Constructs a new [Result] from a function that might throw
  static Result<Success, Failure>
      tryCatch<Success, Failure, Error extends Object>(
    Success Function() tryFn,
    Failure Function(Error error) onError,
  ) {
    try {
      return Ok(tryFn());
    } on Error catch (e) {
      return Err(onError(e));
    }
  }

  /// Constructs a new [Result] from an **async** function that might throw
  static Future<Result<Success, Failure>>
      tryCatchAsync<Success, Failure, Error extends Object>(
    Future<Success> Function() tryFn,
    Failure Function(Error error) onError,
  ) async {
    try {
      return Ok(await tryFn());
    } on Error catch (e) {
      return Err(onError(e));
    }
  }

  /// If the condition is satify then return [value] in
  /// [Ok] else [error] in [Err]
  static Result<Success, Failure> conditional<Success, Failure>(
    Success value,
    Failure error, {
    required bool test,
  }) =>
      test ? Ok(value) : Err(error);

  /// If the condition is satify then return *command* [value]
  /// in [Ok] else [error] in [Err]
  static Result<Success, Failure> conditionalLazy<Success, Failure>(
    Success Function() value,
    Failure Function() error,
    bool Function() predicate,
  ) =>
      predicate.call() ? Ok(value()) : Err(error());

  @override
  int get hashCode => _result.hashCode;

  @override
  bool operator ==(Object other) {
    if (identical(this, other)) {
      return true;
    }
    return other is Result<Success, Failure> && other._result == _result;
  }
}

final class Ok<Success, Failure> extends Result<Success, Failure> {
  Ok(this.ok) : super._(sealed.Ok(ok));

  @override
  final Success ok;

  @override
  bool get isErr => false;

  @override
  bool get isOk => true;

  @override
  Failure? get err => null;

  @override
  String toString() => switch (ok) {
        null => 'Ok(null)',
        Future() => 'Ok(${ok.runtimeType})',
        _ => 'Ok($ok)',
      };
}

final class Err<Success, Failure> extends Result<Success, Failure> {
  Err(this.err) : super._(sealed.Err(err));

  @override
  final Failure err;

  @override
  bool get isErr => true;

  @override
  bool get isOk => false;

  @override
  Success? get ok => null;

  @override
  String toString() => switch (err) {
        _ => 'Err($err)',
      };
}
