// Copyright (C) 2023 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:flutter/widgets.dart';

/// {@template rich_text_style_parameter}
/// Rich text style parameter used to parse the text.
/// {@endtemplate}
class RichTextStyleParameter {
  /// {@macro rich_text_style_parameter}
  const RichTextStyleParameter(
    this.defaultStyle,
    this.definedStyle,
    this.styleName,
  );

  /// Default style to use if no style is defined.
  final TextStyle? defaultStyle;

  /// Map of defined styles. The key is the style name.
  ///
  /// Example:
  /// ```dart
  /// definedStyle: {
  ///  'bold': TextStyle(fontWeight: FontWeight.bold),
  ///  'italic': TextStyle(fontStyle: FontStyle.italic),
  /// }
  /// ```
  /// then, the text `This is <bold>bold</bold> and <italic>italic</italic>`
  /// will be parsed and the word `bold` will be bold and the word `italic`
  /// will be italic in the RichText widget.
  final Map<String, TextStyle> definedStyle;

  /// Style name to use for the text.
  /// If no style is defined and no default style is defined, the text will be
  /// displayed with the default style of the RichText widget.
  final String? styleName;

  /// Returns the style to use for the given style name.
  /// If no style is defined for the given style name, the default style is
  /// returned. If no default style is defined, the default style of the
  /// RichText widget is returned.
  TextStyle get style {
    if (definedStyle.containsKey(styleName)) {
      return definedStyle[styleName]!;
    }
    return defaultStyle ?? const TextStyle();
  }

  RichTextStyleParameter copyWith({
    TextStyle? defaultStyle,
    Map<String, TextStyle>? definedStyle,
    String? styleName,
  }) =>
      RichTextStyleParameter(
        defaultStyle ?? this.defaultStyle,
        definedStyle ?? this.definedStyle,
        styleName ?? this.styleName,
      );
}

/// {@template rich_text_node}
/// Rich text node.
///
/// Used as a node in the tree of the RichText widget.
/// {@endtemplate}
class RichTextNode {
  /// {@macro rich_text_node}
  const RichTextNode(this.nodes);

  /// List of nodes.
  final List<RichTextNode> nodes;

  /// Returns a RichTextNode from the given content.
  /// If the content contains a match with the given regex, the content is
  /// split into multiple nodes. If the content does not contain a match with
  /// the given regex, a leaf node is returned.
  ///
  /// Example:
  /// ```dart
  /// RichTextNode.from(
  ///  'This is <bold>bold</bold> and <italic>italic</italic>',
  ///   RegExp(r'<(\w+)>(.*?)</\1>'),
  ///   RichTextStyleParameter(
  ///     TextStyle(fontWeight: FontWeight.w400),
  ///     {'bold': TextStyle(fontWeight: FontWeight.bold)},
  ///     {'italic': TextStyle(fontStyle: FontStyle.italic)},
  ///   ),
  /// );
  /// ```
  /// will return a RichTextNode with 5 nodes:
  /// - `This is `
  /// - `<bold>bold</bold>`
  /// - ` and `
  /// - `<italic>italic</italic>`
  /// - `` (empty string)
  ///
  /// The first node is a leaf node with the default style of the RichText
  /// widget. The second node is a node with bold paramters. The third node is
  /// a leaf node with the default style of the RichText widget.
  /// The fourth node is a node italic parameters. The fifth node is a
  /// leaf node with the default style of the RichText widget.
  static RichTextNode from(
    String content,
    RegExp regex,
    RichTextStyleParameter styleParameter,
  ) {
    final matches = regex.allMatches(content);
    if (matches.isNotEmpty) {
      // match found -> construct node with leaf/nodes
      final List<RichTextNode> nodes = [];
      for (var i = 0; i < matches.length; i++) {
        final previousMatch = i > 0 ? matches.elementAt(i - 1) : null;
        final currentMatch = matches.elementAt(i);
        // non match before
        final nonMatchBefore = (previousMatch != null)
            ? content.substring(previousMatch.end, currentMatch.start)
            : content.substring(0, currentMatch.start);
        nodes
          ..add(RichTextNode.from(nonMatchBefore, regex, styleParameter))
          // match
          ..add(
            RichTextNode.from(
              currentMatch.group(2)!,
              regex,
              styleParameter.copyWith(styleName: currentMatch.group(1)),
            ),
          );
      }
      // non match after
      final nonMatchAfter = content.substring(matches.last.end);
      nodes.add(RichTextNode.from(nonMatchAfter, regex, styleParameter));
      return RichTextNode(nodes);
    } else {
      // match not found -> construct leaf
      return RichTextLeaf(styleParameter.style, content);
    }
  }

  /// Returns an InlineSpan from the given RichTextNode.
  /// The given RichTextParser is used to convert the RichTextNode to an
  /// InlineSpan.
  ///
  /// InlineSpan is used to display text in the RichText widget.
  InlineSpan toInlineSpan(RichTextParser parser) {
    final children = <InlineSpan>[];
    for (final node in nodes) {
      children.add(node.toInlineSpan(parser));
    }
    return TextSpan(children: children);
  }
}

/// {@template rich_text_leaf}
/// Rich text leaf.
/// {@endtemplate}
class RichTextLeaf extends RichTextNode {
  /// {@macro rich_text_leaf}
  const RichTextLeaf(this.style, this.content) : super(const []);

  /// Style to use for the text.
  final TextStyle style;

  /// Text content.
  final String content;

  @override
  InlineSpan toInlineSpan(RichTextParser parser) =>
      parser.nodeBuilder.call(content, style);
}

/// {@template rich_text_parser}
/// Rich text parser.
/// {@endtemplate}
class RichTextParser {
  /// {@macro rich_text_parser}
  const RichTextParser({required this.nodeBuilder});

  /// Returns a default RichTextParser.
  /// The default RichTextParser uses the given nodeBuilder to convert a
  /// RichTextNode to an InlineSpan.
  ///
  /// By default, the nodeBuilder returns a TextSpan with the given content
  /// and style.
  factory RichTextParser.defaultBuilder() => RichTextParser(
        nodeBuilder: (content, style) => TextSpan(
          text: content,
          style: style,
        ),
      );

  /// Converts the given RichTextNode to an InlineSpan.
  final InlineSpan Function(String content, TextStyle style) nodeBuilder;
}
